import json
import os
import tempfile
from collections import defaultdict
from pprint import pformat

import pydash
from copy import deepcopy
from jsonschema import validate

from infra_buddy.aws.cloudformation import CloudFormationBuddy
from infra_buddy.aws.s3 import S3Buddy, CloudFormationDeployS3Buddy
from infra_buddy.deploy.deploy import Deploy
from infra_buddy.utility import helper_functions, print_utility

_PARAM_TYPE_PROPERTY = "property"
_PARAM_TYPE_TRANSFORM = "transform"

_PARAM_TYPE_FUNC = "func"

_PARAM_TYPE_TEMPLATE = "template"


class CloudFormationDeploy(Deploy):
    schema = {
        "type": "object",
        "additionalProperties": {
            "type": "object",
            "properties": {
                "type": {"type": "string", "enum": [_PARAM_TYPE_PROPERTY,
                                                    _PARAM_TYPE_FUNC,
                                                    _PARAM_TYPE_TEMPLATE,
                                                    _PARAM_TYPE_TRANSFORM]},
                "value": {"type": "string"},
                "default_value": {"type": "string"},
                "key": {"type": "string"}
            },
            "required": ["type"]
        }

    }

    def __init__(self, stack_name, template, deploy_ctx):
        # type: (str, Template,DeployContext) -> None
        super(CloudFormationDeploy, self).__init__(deploy_ctx)
        self.stack_name = stack_name
        self.config_directory = template.get_config_dir()
        self.parameter_file = template.get_parameter_file_path()
        self.template_file = template.get_template_file_path()
        self.default_path = template.get_defaults_file_path()
        self._load_defaults(template.get_default_env_values())

    def _load_defaults(self, default_env_values):
        self.defaults = {}
        if self.default_path and os.path.exists(self.default_path):
            with open(self.default_path, 'r') as default_fp:
                def_obj = json.load(default_fp)
                validate(def_obj, self.schema)
            self._process_default_dict(def_obj)
        self.defaults.update(default_env_values)

    def _process_default_dict(self, def_obj):
        _transformations = {}
        for key, value in def_obj.iteritems():
            self.defaults[key] = self._load_value(value, key, _transformations)
        for key, value in _transformations.iteritems():
            transform = self.transform(value,self.defaults.get(key,None))
            if transform is not None:
                self.defaults[key] = transform

    def _load_value(self, value, key, transformations):
        type_ = value['type']
        if type_ == _PARAM_TYPE_TEMPLATE:
            return self.deploy_ctx.expandvars(value['value'], self.defaults)
        elif type_ == _PARAM_TYPE_FUNC:
            if 'default_key' in value:
                # look for a default value before calling the func
                def_key = value['default_key']
                if def_key in self.defaults:
                    return self.defaults[def_key]
                elif def_key in self.deploy_ctx:
                    return self.deploy_ctx[def_key]
            # someday make it dynamic
            func_name = value['func_name']
            if "load_balancer_name" == func_name:
                return helper_functions.load_balancer_name(self.deploy_ctx)
            elif 'rule_priority' == func_name:
                return helper_functions.calculate_rule_priority(self.deploy_ctx, self.stack_name)
            else:
                print_utility.error(
                    "Can not locate function for defaults.json: Stack {} Function {}".format(self.stack_name,
                                                                                             func_name))
        elif type_ == _PARAM_TYPE_PROPERTY:
            default_value = value.get('default', None)
            if isinstance(default_value, basestring):
                default_value = self.deploy_ctx.expandvars(str(default_value), self.defaults)
            return self.deploy_ctx.get(value['key'], default_value)
        elif type_ == _PARAM_TYPE_TRANSFORM:
            # add it to the list of properties to transform after load
            transformations[key] = value
            # Load like a normal property, so override the type
            value['type'] = _PARAM_TYPE_PROPERTY
            # and recurse
            return self._load_value(value, None, None)
        else:
            # should die on JSON validation but to be complete
            print_utility.error(
                "Can not load value for type in defaults.json: Stack {} Type {}".format(self.stack_name, type_))

    def transform(self, definition, value):
        func_name = definition['func_name']
        if 'transform_fargate_cpu' == func_name:
            return helper_functions.transform_fargate_cpu(self.defaults, value)
        elif 'transform_fargate_memory' == func_name:
            return helper_functions.transform_fargate_memory(self.defaults, value)
        else:
            print_utility.error(
                "Can not locate function for defaults.json: Stack {} Function {}".format(self.stack_name,
                                                                                         func_name))

    def get_rendered_config_files(self):
        self._prep_render_destination()
        rendered_config_files = []
        config_dir = self.config_directory
        if config_dir:
            for template in os.listdir(config_dir):
                rendered_config_files.append(
                    self.deploy_ctx.render_template(os.path.join(config_dir, template), self.destination))
        return rendered_config_files

    def get_rendered_param_file(self):
        self._prep_render_destination()
        return self.deploy_ctx.render_template(self.parameter_file, self.destination)

    def validate(self):
        self.print_template_description()
        self.print_known_parameters()
        self.print_export()
        config_files = self.get_rendered_config_files()
        if len(config_files) == 0:
            print_utility.warn("No Configuration Files")
        else:
            print_utility.warn("Configuration Files:")
            for config_file in config_files:
                self._print_file(config_file)

    def _print_file(self, config_file):
        with open(config_file, 'r') as cf:
            print_utility.warn(os.path.basename(config_file))
            for line in cf.readlines():
                print_utility.banner(line)

    def _prep_render_destination(self):
        self.destination = tempfile.mkdtemp()

    def print_known_parameters(self):
        # type: (DeployContext) -> int
        known_param, warnings, errors = self._analyze_parameters()
        print_utility.banner_warn("Parameters", pformat(known_param))
        print_utility.warn("Parameter Warnings")
        self._print_info(warnings)
        print_utility.warn("Parameter Errors")
        self._print_error(errors)
        return len(errors)

    def print_export(self):
        # type: () -> int
        known_exports, warnings, errors = self._analyze_export()
        print_utility.warn("Export Values")
        self._print_info(known_exports)
        print_utility.warn("Export Values Warnings")
        self._print_error(errors)
        return len(errors)

    def analyze(self):
        errs = self.print_known_parameters()
        errs += self.print_export()
        return errs

    def _analyze_parameters(self):
        known_param = {}
        errors = defaultdict(list)
        warning = defaultdict(list)
        ssm_keys = []
        with open(self.template_file, 'r') as template:
            template_obj = json.load(template)
            template_params = pydash.get(template_obj, 'Parameters', {})
            for key, value in template_params.iteritems():
                description = value.get('Description', None)
                default = value.get('Default', None)
                if not description: warning[key].append("Parameter does not contain a description")
                if default:
                    type = value.get('Type',None)
                    if type:
                        if type and 'AWS::SSM' in type:
                            ssm_keys.append(key)
                    else:
                        warning[key].append("Parameter has default value defined in CloudFormation Template - {}".format(default))
                known_param[key] = {'description': description, 'type': value['Type']}
        value_to_key = {}
        with open(self.parameter_file, 'r') as params:
            param_file_params = json.load(params)
            for param in param_file_params:
                key_ = param['ParameterKey']
                if key_ in known_param:
                    known_param[key_]['variable'] = param['ParameterValue']
                    value_to_key[param['ParameterValue'].replace("$", "").replace("{", "").replace("}", "")] = key_
                    expandvars = self.deploy_ctx.expandvars(param['ParameterValue'], self.defaults)
                    if "${" in expandvars: warning[key_].append(
                        "Parameter did not appear to validate ensure it is populated when using the template - {}"
                            .format(expandvars))
                    known_param[key_]['default_value'] = expandvars
                else:
                    if key_ not in ssm_keys:
                        # exists in param file but not in template
                        errors[key_].append("Parameter does not exist in template but defined in param file")
        if self.default_path and os.path.exists(self.default_path):
            with open(self.default_path, 'r') as defs:
                defs = json.load(defs)
                for key_, param in defs.iteritems():
                    if key_ in value_to_key:
                        param_key = value_to_key[key_]
                        known_param[param_key]['default_type'] = param['type']
                        known_param[param_key].update(param)
                    else:
                        # exists in param file but not in template
                        errors[key_].append("Parameter does not exist in parameter file but defined in defaults file")

        for key, value in known_param.iteritems():
            if 'variable' not in value:
                errors[key].append("Parameter does not exist in param file but defined in template")
        return known_param, warning, errors

    def _analyze_export(self):
        known_exports = {}
        errors = defaultdict(list)
        warnings = defaultdict(list)
        with open(self.template_file, 'r') as template:
            template_obj = json.load(template)
            template_exports = pydash.get(template_obj, 'Outputs', {})
            for key, value in template_exports.iteritems():
                export = value.get('Export', None)
                value = value.get('Value', None)
                description = value.get('Description', None)
                if not description: warnings[key].append("Export does not contain a description")
                known_exports[key] = {'description': description, 'export': export, 'value': value}
        return known_exports, warnings, errors

    def print_template_description(self):
        with open(self.template_file, 'r') as template:
            template_obj = json.load(template)
            print_utility.banner_warn("Deploy for Stack: {}".format(self.stack_name),
                                      pydash.get(template_obj, 'Description', ''))

    def _print_error(self, errors):
        for key, errs in errors.iteritems():
            print_utility.error(pformat(key, indent=4))
            print_utility.banner(pformat(errs, indent=8))

    def _print_info(self, errors):
        for key, errs in errors.iteritems():
            print_utility.warn(pformat(key, indent=4))
            print_utility.banner(pformat(errs, indent=8))

    def _internal_deploy(self, dry_run):
        # Initialize our buddies
        s3 = CloudFormationDeployS3Buddy(self.deploy_ctx)
        cloud_formation = CloudFormationBuddy(self.deploy_ctx)
        if dry_run:
            self.validate()
            return
        # Upload our template to s3 to make things a bit easier and keep a record
        template_file_url = s3.upload(file=(self.template_file))
        # Upload all of our config files to S3 rendering any variables
        config_files = self.get_rendered_config_files()
        for rendered in config_files:
            s3.upload(file=rendered)
        # render our parameter files
        parameter_file_rendered = self.get_rendered_param_file()
        # see if we are updating or creating
        if cloud_formation.should_create_change_set():
            cloud_formation.create_change_set(template_file_url=template_file_url,
                                              parameter_file=parameter_file_rendered)
            # make sure it is available and that there are no special conditions
            if cloud_formation.should_execute_change_set():
                print_utility.progress("Updating existing stack with ChangeSet - {}".format(self.stack_name))
                cloud_formation.execute_change_set()
            else:
                print_utility.warn("No computed changes for stack - {}".format(self.stack_name))
                # if there are no changes then clean up and exit
                cloud_formation.delete_change_set()
                return
        else:
            print_utility.progress("Creating new stack - {}".format(self.stack_name))
            cloud_formation.create_stack(template_file_url=template_file_url,
                                         parameter_file=parameter_file_rendered)

    def get_default_params(self):
        return self._analyze_parameters()[0]



import math
from operator import itemgetter

import boto3

from infra_buddy.aws.cloudformation import CloudFormationBuddy
from infra_buddy.utility import print_utility


def _get_all_export_values(client):
    pass


def load_balancer_name(deploy_ctx):
    # type: (DeployContext) -> str
    cf = CloudFormationBuddy(deploy_ctx)
    # we need an export value from the cluster stack so manually override it
    val = _get_cluster_stack_export_value(cf, deploy_ctx, "ElasticLoadBalancerARN")
    # //#amazon conveniently wants some substring of the ARN instead of the name or other value actually available in the API
    # //# turn arn:aws:elasticloadbalancing:us-west-2:271083817914:listener/app/prod-EcsEl-1WYNMMT2MT9NR/c5f92ddeb151227f/313bb2e23d9dd8d8
    # //# into app/prod-EcsEl-1WYNMMT2MT9NR/c5f92ddeb151227f/313bb2e23d9dd8d8
    return "" if not val else val[val.find('app/'):]


def _get_cluster_stack_export_value(cf, deploy_ctx, param):
    # type: (CloudFormationBuddy,DeployContext) -> str
    val = None
    try:
        cf.stack_name = deploy_ctx.cluster_stack_name
        val = cf.get_export_value(param)
    except Exception as e:
        print_utility.warn("Exception getting export for helper function - {}".format(e))
    finally:
        return val


def _get_max_priority(rules):
    rules = sorted(rules, key=lambda k: k["Priority"], reverse=True)
    for rule in rules:
        priority_ = rule['Priority']
        if priority_ != "default":
            return int(priority_)


# 0.5GB, 1GB, 2GB - Available cpu values: 256 (.25 vCPU)
# 1GB, 2GB, 3GB, 4GB - Available cpu values: 512 (.5 vCPU)
# 2GB, 3GB, 4GB, 5GB, 6GB, 7GB, 8GB - Available cpu values: 1024 (1 vCPU)
# Between 4GB and 16GB in 1GB increments - Available cpu values: 2048 (2 vCPU)
# Between 8GB and 30GB in 1GB increments - Available cpu values: 4096 (4 vCPU)
def _get_valid_fargate_memory(_value):
    if _value <= 512:
        print_utility.info("Transforming memory value of {} to '0.5GB' - min value".format(_value))
        return '512'
    elif _value < 1024:
        print_utility.info("Transforming memory value of {} to '1GB' - next legitimate value".format(_value))
        return '1024'
    else:
        memory = int(math.ceil(_value/1024.0))
        memory = memory * 1024 #normalize to closest interval of 1GB
        if memory > 30720:
            print_utility.info("Transforming memory value of {} to '30GB' - max value".format(_value))
            memory = 30720
        else:
            print_utility.info("Transforming memory value of {} to '{}GB'".format(_value, memory))
        return memory


def _get_valid_fargate_cpu(_value):
    if _value <= 256:
        print_utility.info("Transforming cpu value of {} to '256' - min value".format(_value))
        return 256
    elif _value <= 512:
        print_utility.info("Transforming cpu value of {} to '512' - next valid value".format(_value))
        return 512
    elif _value <= 1024:
        print_utility.info("Transforming cpu value of {} to '1024' - next valid value".format(_value))
        return 1024
    elif _value <= 2048:
        print_utility.info("Transforming cpu value of {} to '2048' - next valid value".format(_value))
        return 2048
    elif _value <= 4096:
        print_utility.info("Transforming cpu value of {} to '4096' - next valid value".format(_value))
        return 4096
    else:
        print_utility.info("Transforming cpu value of {} to '4096' - max valid value".format(_value))
        return 4096


# 256 (.25 vCPU) - Available memory values: 0.5GB, 1GB, 2GB
# 512 (.5 vCPU) - Available memory values: 1GB, 2GB, 3GB, 4GB
# 1024 (1 vCPU) - Available memory values: 2GB, 3GB, 4GB, 5GB, 6GB, 7GB, 8GB
# 2048 (2 vCPU) - Available memory values: Between 4GB and 16GB in 1GB increments
# 4096 (4 vCPU) - Available memory values: Between 8GB and 30GB in 1GB increments
_valid_fargate_resources = {
    256: ['512', '1024', '2048'],
    512: ["{}".format(i*1024) for i in range(1, 4)],
    1024: ["{}".format(i*1024) for i in range(2, 8)],
    2048: ["{}".format(i*1024) for i in range(4, 16)],
    4096: ["{}".format(i*1024) for i in range(8, 30)],
}

_valid_fargate_memories = set([item for sublist in _valid_fargate_resources.itervalues() for item in sublist])



def _validate_fargate_resource_allocation(cpu, memory, deploy_ctx):
    if cpu is None:
        discovered_cpu = deploy_ctx.get('TASK_CPU', None)
        if discovered_cpu not in _valid_fargate_resources:
            print_utility.info('Skipping fargate resource validation - CPU not transformed - {}'.format(discovered_cpu))
            return
        cpu = discovered_cpu
    elif memory is None:
        discovered_memory = deploy_ctx.get('TASK_SOFT_MEMORY', None)
        if discovered_memory not in _valid_fargate_memories:
            print_utility.info(
                'Skipping fargate resource validation - Memory not transformed - {}'.format(discovered_memory))
            return
        memory = discovered_memory
    memory_possibilities = _valid_fargate_resources[cpu]
    if memory not in memory_possibilities:
        print_utility.error(
            'Attempting to use fargate with invalid configuration.  {} CPU {} Memory'.format(cpu, memory),
            raise_exception=True)


def transform_fargate_cpu(deploy_ctx, _value):
    if _using_fargate(deploy_ctx):
        cpu = _get_valid_fargate_cpu(_value)
        _validate_fargate_resource_allocation(cpu, None, deploy_ctx)
        return cpu


def _using_fargate(deploy_ctx):
    return deploy_ctx.get('USE_FARGATE', 'false') == 'true'


def transform_fargate_memory(deploy_ctx, _value):
    if _using_fargate(deploy_ctx):
        if isinstance(_value, basestring):
            if _value not in _valid_fargate_memories:
                print_utility.error(
                    'Attempting to use fargate with invalid memory.  {} Memory Valid Values: {}'.format(_value,
                                                                                                        _valid_fargate_memories),
                    raise_exception=True)
            else:
                memory = _value
        else:
            memory = "{}".format(_get_valid_fargate_memory(_value))
        _validate_fargate_resource_allocation(None, memory, deploy_ctx)
        return memory


def calculate_rule_priority(deploy_ctx, stack_name):
    # type: (DeployContext,str) -> str
    cf = CloudFormationBuddy(deploy_ctx)
    # we need some data for the passed stack_name so manually override it
    cf.stack_name = stack_name
    if cf.does_stack_exist():
        return cf.get_existing_parameter_value('RulePriority')
    else:
        listenerArn = _get_cluster_stack_export_value(cf, deploy_ctx, "ListenerARN")
        if listenerArn:
            client = get_boto_client(deploy_ctx)
            rules = client.describe_rules(ListenerArn=listenerArn)['Rules']
        else:
            rules = None
        if not rules or len(rules) == 0:
            current_max = 30
        else:
            if len(rules) == 1 and rules[0]['Priority'] == "default":
                current_max = 30
            else:
                current_max = int(_get_max_priority(rules))
        return str(current_max + 1)

    # //if [[ $(does_stack_exist ${STACK_NAME}) == "Yes" ]]; then
    # //    RULE_PRIORITY=$(print_stack_param ${STACK_NAME} "RulePriority")
    # //else
    # //    ListenerARN=`print_export_value "$(generate_cluster_stack_name)-ListenerARN"`
    # //    EXISTING_RULES=`aws elbv2 describe-rules --listener-arn ${ListenerARN} | jq ".Rules"`
    # //    if [[ ${EXISTING_RULES} != "null" ]]; then
    # //        CURRENT_MAX=$(aws elbv2 describe-rules --listener-arn ${ListenerARN} | jq ".Rules[].Priority" | sed 's/[^0-9]*//g' | sort -nr | head -n1)
    # //    else
    # //        CURRENT_MAX=30
    # //    fi
    # //    #increment by 30 to allow plenty of room for explicit rule creation
    # //    CURRENT_NUMBER_OF_RULES=$((CURRENT_MAX+1))
    # //fi


def get_boto_client(deploy_ctx):
    return boto3.client('elbv2', region_name=deploy_ctx.region)



import json
import os

import boto3
import click
from click import UsageError

from infra_buddy.commandline import cli
from infra_buddy.context.deploy_ctx import DeployContext
from infra_buddy.deploy.cloudformation_deploy import CloudFormationDeploy
from infra_buddy.template.template import NamedLocalTemplate
from infra_buddy.utility import print_utility


@cli.command(name='bootstrap', short_help="Generate keys for a new environment to be managed by infra-buddy.")
@click.argument('environments', nargs=-1)
@click.pass_obj
def deploy_cloudformation(deploy_ctx, environments):
    # type: (DeployContext,list) -> None
    do_command(deploy_ctx, environments)


def do_command(deploy_ctx, environments, destination=None):
    # type: (DeployContext,list) -> None
    client = boto3.client('ec2', region_name=deploy_ctx.region)
    if len(environments) == 0:
        raise UsageError("Expected at least one environment (ci, prod)")
    for env in environments:
        key_name = "{env}-{application}".format(env=env, application=deploy_ctx.application)
        res = client.create_key_pair(KeyName=key_name)
        key_location = '{key_name}.pem'.format(key_name=key_name)
        if destination:
            key_location = os.path.join(destination, key_location)
        with open(key_location, 'w') as new_pem:
            new_pem.writelines(res['KeyMaterial'])
    def_obj = {
        'service-templates': {},
        'service-modification-templates': {}
    }
    defaults = "defaults.json"
    if destination:
        defaults = os.path.join(destination, defaults)
    with open(defaults, 'w') as default_file:
        json.dump(def_obj, default_file)



import boto3
import pydash
from copy import deepcopy

from infra_buddy.utility import print_utility, waitfor

from infra_buddy.aws.cloudformation import CloudFormationBuddy


class ECSBuddy(object):
    def __init__(self, deploy_ctx):
        # type: (DeployContext) -> None
        super(ECSBuddy, self).__init__()
        self.deploy_ctx = deploy_ctx
        self.client = boto3.client('ecs', region_name=self.deploy_ctx.region)
        cf = CloudFormationBuddy(deploy_ctx)
        ecs_cluster_export_key = "{}-ECSCluster".format(self.deploy_ctx.cluster_stack_name)
        self.cluster = self._wait_for_export(cf=cf, fully_qualified_param_name=ecs_cluster_export_key)
        ecs_service_export_key = "{}-ECSService".format(self.deploy_ctx.stack_name)
        self.ecs_service = self._wait_for_export(cf=cf, fully_qualified_param_name=ecs_service_export_key)
        ecs_task_family_export_key = "{}-ECSTaskFamily".format(self.deploy_ctx.stack_name)
        self.ecs_task_family = self._wait_for_export(cf=cf, fully_qualified_param_name=ecs_task_family_export_key)
        ecs_task_execution_role_export_key = "{}-ECSTaskRole".format(self.deploy_ctx.stack_name)
        self.ecs_task_execution_role = self._wait_for_export(cf=cf, fully_qualified_param_name=ecs_task_execution_role_export_key)
        self.task_definition_description = None
        self.new_image = None

    @classmethod
    def _wait_for_export(cls, cf, fully_qualified_param_name):
        # we are seeing an issue where immediately after stack create the export values are not
        # immediately available
        value = waitfor.waitfor(
            function_pointer=cf.get_export_value,
            expected_result=None,
            interval_seconds=2,
            max_attempts=5,
            negate=True,
            args={"fully_qualified_param_name": fully_qualified_param_name},
            exception=False
        )

        print_utility.info("[wait_for_export] {}={}".format(fully_qualified_param_name, value))
        return value

    def set_container_image(self, location, tag):
        self.new_image = "{location}:{tag}".format(location=location, tag=tag)

    def requires_update(self):
        if not self.new_image:
            print_utility.warn("Checking for ECS update without registering new image ")
            return False
        if not self.ecs_task_family:
            print_utility.warn("No ECS Task family found - assuming first deploy of stack and skipping ECS update")
            return False
        self._describe_task_definition()
        existing = pydash.get(self.task_definition_description, "containerDefinitions[0].image")
        print_utility.info("ECS task existing image - {}".format(existing))
        print_utility.info("ECS task desired image - {}".format(self.new_image))
        return existing != self.new_image

    def perform_update(self):
        self._describe_task_definition(refresh=True)
        new_task_def = {
            'family': self.task_definition_description['family'],
            'containerDefinitions': self.task_definition_description['containerDefinitions'],
            'volumes': self.task_definition_description['volumes']
        }
        if 'networkMode' in self.task_definition_description:
            new_task_def['networkMode'] = self.task_definition_description['networkMode']
        new_task_def['containerDefinitions'][0]['image'] = self.new_image

        using_fargate = self.deploy_ctx.get('USE_FARGATE') == 'true'

        ctx_memory = self.deploy_ctx.get('TASK_MEMORY')
        if ctx_memory:
            new_task_def['containerDefinitions'][0]['memory'] = ctx_memory

        if 'TASK_SOFT_MEMORY' in self.deploy_ctx and self.deploy_ctx['TASK_SOFT_MEMORY']:
            new_task_def['containerDefinitions'][0]['memoryReservation'] = self.deploy_ctx['TASK_SOFT_MEMORY']

        ctx_cpu = self.deploy_ctx.get('TASK_CPU')
        if ctx_cpu:
            new_task_def['containerDefinitions'][0]['cpu'] = ctx_cpu

        # set at the task level for fargate definitions
        if using_fargate:
            first_container = new_task_def['containerDefinitions'][0]
            new_task_def['requiresCompatibilities'] = ['FARGATE']
            new_cpu = ctx_cpu or first_container.get('cpu')
            if new_cpu:
                new_task_def['cpu'] = str(new_cpu)  # not sure if this is right but AWS says it should be str

            new_memory = ctx_memory or first_container.get('memory')
            if new_memory:
                new_task_def['memory'] = str(new_memory)  # not sure if this is right but AWS says it should be str

        if self.ecs_task_execution_role:
            new_task_def['executionRoleArn'] = self.ecs_task_execution_role

        for k, v in self.deploy_ctx.items():
            print_utility.info('[deploy_ctx] {} = {}'.format(k, repr(v)))

        for k, v in new_task_def.items():
            print_utility.info('[new_task_def] {} = {}'.format(k, repr(v)))

        updated_task_definition = self.client.register_task_definition(**new_task_def)['taskDefinition']
        new_task_def_arn = updated_task_definition['taskDefinitionArn']

        self.deploy_ctx.notify_event(
            title="Update of ecs service {service} started".format(service=self.ecs_service),
            type="success")
        self.client.update_service(
            cluster=self.cluster,
            service=self.ecs_service,
            taskDefinition=new_task_def_arn)
        waiter = self.client.get_waiter('services_stable')
        success = True
        try:
            waiter.wait(
                cluster=self.cluster,
                services=[self.ecs_service]
            )
        except Exception as e:
            success = False
            print_utility.error("Error waiting for service to stabilize - {}".format(e.message), raise_exception=True)
        finally:
            self.deploy_ctx.notify_event(
                title="Update of ecs service {service} completed".format(service=self.ecs_service,
                                                                         success="Success" if success else "Failed"),
                type="success" if success else "error")

    def _describe_task_definition(self, refresh=False):
        if self.task_definition_description and not refresh:
            return
        self.task_definition_description = self.client.describe_task_definition(taskDefinition=self.ecs_task_family)[
            'taskDefinition']



import datadog as dd


class DataDogNotifier(object):
    def __init__(self, key, deploy_context):
        # type: (DeployContext) -> None
        super(DataDogNotifier, self).__init__()
        self.deploy_context = deploy_context
        dd.initialize(api_key=key)

    def notify_event(self, title, type, message):
        dd.api.Event.create(title=title,
                            text=message,
                            alert_type=type,
                            tags=self._get_tags())

    def _get_tags(self):
        return ['application:{app},role:{role},environment:{env}'.format(
            app=self.deploy_context.application,
            role=self.deploy_context.role,
            env=self.deploy_context.environment)]



import click

from infra_buddy.commandline import cli
from infra_buddy.context.artifact_definition import ArtifactDefinition
from infra_buddy.context.deploy_ctx import DeployContext
from infra_buddy.utility import print_utility


@cli.command(name='generate-artifact-manifest',
             short_help="Generate an artifact manifest for use by the deploy-service command.")
@click.option("--artifact-type",
              help="The type of artifact referenced in the manifest ( 'container' or 's3' currently supported).")
@click.option("--artifact-location",
              help="The location of the artifact referenced in the manifest (Docker registry or S3 bucket and path).")
@click.option("--artifact-identifier",
              help="The identifier for the artifact. (Docker tag or filename excluding 'zip' extension).")
def deploy_cloudformation(artifact_type, artifact_location, artifact_identifier):
    # type: (str,str,str) -> None
    path  = do_command(artifact_type, artifact_location, artifact_identifier)
    print_utility.info("Artifact Manifest saved to - {}".format(path))




def do_command(artifact_type, artifact_location, artifact_identifier, destination=None):
    # type: (str,str,str) -> str
    ad = ArtifactDefinition.create(artifact_type, artifact_location, artifact_identifier)
    print_utility.info("Generated artifact manifest - {}".format(ad.__class__.__class__))
    return ad.save_to_file(destination_dir=destination)



import click
import time

import datetime

from infra_buddy.commandline import cli
from infra_buddy.context.deploy_ctx import DeployContext
from infra_buddy.utility import print_utility

current_milli_time = lambda: int(round(time.time() * 1000))


@cli.command(name='deploy-service', short_help="Deploy and/or update a service as defined by a service definition "
                                               "and an optional artifact definition.  Operation is idempotent.")
@click.option("--dry-run", is_flag=True, help="Prints the execution plan and displays the evaluated "
                                              "parameter values for the deployment.")
@click.pass_obj
def deploy_cloudformation(deploy_ctx, dry_run):
    # type: (DeployContext,bool) -> None
    do_command(deploy_ctx, dry_run)

def do_command(deploy_ctx, dry_run):
    # type: (DeployContext,bool) -> None
    plan = deploy_ctx.get_execution_plan()
    for deploy in plan:
        start = current_milli_time()
        print_utility.progress("Starting Deployment: {}".format(str(deploy)))
        deploy.do_deploy(dry_run)
        print_utility.progress("Finished Deployment in {} ".format(
            print_utility.print_time_delta(datetime.timedelta(milliseconds=(current_milli_time() - start)))))



import datetime
import json
import os
import re
import tempfile
from collections import OrderedDict
from pprint import pformat

from infra_buddy.aws import s3
from infra_buddy.context.artifact_definition import ArtifactDefinition
from infra_buddy.context.monitor_definition import MonitorDefinition
from infra_buddy.context.service_definition import ServiceDefinition
from infra_buddy.notifier.datadog_notifier import DataDogNotifier
from infra_buddy.template.template_manager import TemplateManager
from infra_buddy.utility import print_utility

STACK_NAME = 'STACK_NAME'

DOCKER_REGISTRY = 'DOCKER_REGISTRY_URL'
ROLE = 'ROLE'
IMAGE = 'IMAGE'
APPLICATION = 'APPLICATION'
ENVIRONMENT = 'ENVIRONMENT'
REGION = 'REGION'
SKIP_ECS = 'SKIP_ECS'
built_in = [DOCKER_REGISTRY, ROLE, APPLICATION, ENVIRONMENT, REGION, SKIP_ECS]
env_variables = OrderedDict()
env_variables['VPCAPP'] = "${VPCAPP}"
env_variables['DEPLOY_DATE'] = "${DEPLOY_DATE}"
env_variables[STACK_NAME] = "${ENVIRONMENT}-${APPLICATION}-${ROLE}"
env_variables['EnvName'] = "${STACK_NAME}"  # alias
env_variables['ECS_SERVICE_STACK_NAME'] = "${STACK_NAME}"  # alias
env_variables['VPC_STACK_NAME'] = "${ENVIRONMENT}-${VPCAPP}-vpc"
env_variables['CF_BUCKET_NAME'] = "${ENVIRONMENT}-${VPCAPP}-cloudformation-deploy-resources"
env_variables['TEMPLATE_BUCKET'] = "${ENVIRONMENT}-${VPCAPP}-cloudformation-deploy-resources" # alias
env_variables['CF_DEPLOY_RESOURCE_PATH'] = "${STACK_NAME}/${DEPLOY_DATE}"
env_variables['CONFIG_TEMPLATES_URL'] = "https://s3-${REGION}.amazonaws.com/${CF_BUCKET_NAME}/${CF_DEPLOY_RESOURCE_PATH}"
env_variables['CONFIG_TEMPLATES_EAST_URL'] = "https://s3.amazonaws.com/${CF_BUCKET_NAME}/${CF_DEPLOY_RESOURCE_PATH}"
env_variables['CLUSTER_STACK_NAME'] = "${ENVIRONMENT}-${APPLICATION}-cluster"
env_variables['RESOURCE_STACK_NAME'] = "${ENVIRONMENT}-${APPLICATION}-${ROLE}-resources"
env_variables['ECS_SERVICE_RESOURCE_STACK_NAME'] = "${RESOURCE_STACK_NAME}"  # alias
env_variables['KEY_NAME'] = "${ENVIRONMENT}-${APPLICATION}"
env_variables['CHANGE_SET_NAME'] = "${STACK_NAME}-deploy-cloudformation-change-set"




class DeployContext(dict):
    def __init__(self, defaults, environment):
        super(DeployContext, self).__init__()
        self.current_deploy = None
        self.temp_files = []
        self._initalize_defaults(defaults,environment)

    @classmethod
    def create_deploy_context_artifact(cls, artifact_directory, environment, defaults=None):
        # type: (str, str) -> DeployContext
        """
        :rtype DeployContext
        :param artifact_directory: Path to directory containing service definition.
                May be a s3 URL pointing at a zip archive
        :param defaults: Path to json file containing default environment settings
        """
        ret = DeployContext(defaults=defaults, environment=environment)
        ret._initialize_artifact_directory(artifact_directory)
        ret._initialize_environment_variables()
        return ret

    @classmethod
    def create_deploy_context(cls, application, role, environment, defaults=None):
        # type: (str, str, str, str) -> DeployContext
        """
        :rtype DeployContext
        :param application: Application name
        :param role: Role of service
        :param environment: Environment to deploy
        :param defaults: Path to json file containing default environment settings
        """
        ret = DeployContext(defaults=defaults, environment=environment)
        ret['APPLICATION'] = application
        ret['ROLE'] = role
        ret._initialize_environment_variables()
        return ret

    def print_self(self):
        print_utility.warn("Context:")
        print_utility.warn("Stack: {}".format(self.stack_name))
        if len(self.stack_name_cache)>0:
            print_utility.warn("Depth: {}".format(self.stack_name_cache))
        if self.current_deploy:
            print_utility.banner_warn("Deploy Defaults:",pformat(self.current_deploy.defaults))
        print_utility.banner_warn("Environment:",pformat(self))

    def _initialize_artifact_directory(self, artifact_directory):
        # type: (str) -> None
        if artifact_directory.startswith("s3://"):
            tmp_dir = tempfile.mkdtemp()
            s3.download_zip_from_s3_url(artifact_directory, destination=tmp_dir)
            artifact_directory = tmp_dir
        service_definition = ServiceDefinition(artifact_directory, self['ENVIRONMENT'])
        self[APPLICATION] = service_definition.application
        self[ROLE] = service_definition.role
        self[DOCKER_REGISTRY] = service_definition.docker_registry
        self.update(service_definition.deployment_parameters)
        self.service_definition = service_definition
        self.artifact_definition = ArtifactDefinition.create_from_directory(artifact_directory)
        self.monitor_definition = MonitorDefinition.create_from_directory(artifact_directory)
        self.artifact_definition.register_env_variables(self)

    def _initialize_environment_variables(self):
        application = self['APPLICATION']
        self['VPCAPP'] = application if not application or '-' not in application else application[:application.find('-')]
        # allow for partial stack names for validation and introspection usecases
        stack_template = "${ENVIRONMENT}"
        if application:
            stack_template += "-${APPLICATION}"
            if self['ROLE']:
                stack_template += "-${ROLE}"
        env_variables[STACK_NAME] = stack_template
        self['DEPLOY_DATE'] = datetime.datetime.now().strftime("%b_%d_%Y_Time_%H_%M")
        for property_name in built_in:
            self.__dict__[property_name.lower()] = self.get(property_name, None)
        for variable, template in env_variables.iteritems():
            evaluated_template = self.expandvars(template)
            self[variable] = evaluated_template
            self.__dict__[variable.lower()] = evaluated_template
        #s3 has non-standardized behavior in us-east-1 you can not use the region in the url
        if self['REGION'] == 'us-east-1':
            self['CONFIG_TEMPLATES_URL'] = self['CONFIG_TEMPLATES_EAST_URL']
            self.__dict__['CONFIG_TEMPLATES_URL'.lower()] = self['CONFIG_TEMPLATES_EAST_URL']

        print_utility.info("deploy_ctx = {}".format(repr(self.__dict__)))

    def _initalize_defaults(self, defaults,environment):
        self['DATADOG_KEY'] = ""
        self['ENVIRONMENT'] = environment.lower() if environment else "dev"
        if defaults:
            print_utility.info("Loading default settings from path: {}".format(defaults))
            with open(defaults, 'r') as fp:
                config = json.load(fp)
                self.update(config)
        self.update(os.environ)
        if 'REGION' not in self:
            print_utility.warn("Region not configured using default 'us-west-1'. "
                               "This is probably not what you want - N. California is slow, like real slow."
                               "  Set the environment variable 'REGION' or pass a default configuration file to override. ")
            self['REGION'] = 'us-west-1'
        self.template_manager = TemplateManager(self.get_deploy_templates(),self.get_service_modification_templates())
        self.stack_name_cache = []
        if self.get('DATADOG_KEY','') is not '':
            self.notifier = DataDogNotifier(key=self['DATADOG_KEY'],deploy_context=self)
        else:
            self.notifier = None

    def get_deploy_templates(self):
        return self.get('service-templates', {})

    def get_service_modification_templates(self):
        return self.get('service-modification-templates', {})

    def generate_modification_stack_name(self, mod_name):
        return "{ENVIRONMENT}-{APPLICATION}-{ROLE}-{mod_name}".format(mod_name=mod_name, **self)

    def generate_modification_resource_stack_name(self, mod_name):
        return "{ENVIRONMENT}-{APPLICATION}-{ROLE}-{mod_name}-resources".format(mod_name=mod_name, **self)

    def get_region(self):
        return self._get_required_default_configuration(REGION)

    def _get_required_default_configuration(self, key):
        region = self.get(key, os.environ.get(key, None))
        if not region:
            raise Exception("Required default not set {key}.\n"
                            "Configure --configuration-defaults or set ENVIRONMENT variable {key}".format(
                key=key))
        return region

    def notify_event(self, title, type, message=None):
        if self.notifier:
            self.notifier.notify_event(title,type,message)
        else:
            print_utility.warn("Notify {type}: {title} - {message}".format(type=type,title=title,message=message))

    def get_service_modifications(self):
        return self.service_definition.service_modifications

    def should_skip_ecs_trivial_update(self):
        return self.get(SKIP_ECS, os.environ.get(SKIP_ECS, "True")) == "True"

    def render_template(self, file,destination):
        with open(file, 'r') as source:
            with open(os.path.join(destination,os.path.basename(file).replace('.tmpl','')),'w+') as destination:
                temp_file_path = os.path.abspath(destination.name)
                print_utility.info("Rendering template to path: {}".format(temp_file_path))
                self.temp_files.append(temp_file_path)
                for line in source:
                    destination.write(self.expandvars(line))
                return temp_file_path

    def __del__(self):
        for file in self.temp_files:
            os.remove(file)

    def get_execution_plan(self):
        # type: () -> list(Deploy)
        execution_plan = self.service_definition.generate_execution_plan(self.template_manager, self)
        artifact_plan = self.artifact_definition.generate_execution_plan(self)
        if artifact_plan:
            execution_plan.extend(artifact_plan)
        monitor_plan = self.monitor_definition.generate_execution_plan(self)
        if monitor_plan:
            execution_plan.extend(monitor_plan)
        print_utility.progress("Execution Plan:")
        for deploy in execution_plan:
            print_utility.info_banner("\t"+str(deploy))
        return execution_plan

    def expandvars(self, template_string, aux_dict=None):
        if not template_string: return template_string #if you pass none, return none
        """Expand ENVIRONMENT variables of form $var and ${var}.
        """
        def replace_var(m):
            if aux_dict:
                val = aux_dict.get(m.group(2) or m.group(1), None)
                if val is not None:return transform(val)
            # if we are in a deployment values set in that context take precedent
            if self.current_deploy is not None:
                val = self.current_deploy.defaults.get(m.group(2) or m.group(1), None)
                if val is not None:return transform(val)
            return transform(self.get(m.group(2) or m.group(1), m.group(0)))

        def transform( val):
            if isinstance(val, bool):
                return str(val).lower()
            return str(val)

        reVar = r'(?<!\\)\$(\w+|\{([^}]*)\})'
        sub = re.sub(reVar, replace_var, template_string)
        return sub

    def recursive_expand_vars(self,source):
        if isinstance(source,dict):
            ret = {}
            for key,value in source.iteritems():
                ret[key] = self.recursive_expand_vars(value)
            return ret
        elif isinstance(source, list):
            ret = []
            for item in source:
                ret.append(self.recursive_expand_vars(item))
            return ret
        elif isinstance(source,basestring):
            return self.expandvars(source)
        else:
            return source


    def push_deploy_ctx(self, deploy_):
        # type: (CloudFormationDeploy) -> None
        if deploy_.stack_name:
            self.stack_name_cache.append(self[STACK_NAME])
            self._update_stack_name(deploy_.stack_name)
        self.current_deploy = deploy_

    def _update_stack_name(self, new_val):
        self[STACK_NAME] = new_val
        self.stack_name = new_val

    def pop_deploy_ctx(self):
        if self.current_deploy.stack_name:
            new_val = self.stack_name_cache.pop()
            self._update_stack_name(new_val)
        self.current_deploy = None

